@@ -19,7 +19,10 @@ Abstract base classes for implementation of protocol specific providers.
 
 """
 
+from authomatic.core import Session
+from authomatic.exceptions import ConfigError, AuthenticationError, FetchError, CredentialsError
 import abc
+import authomatic.core
 import base64
 import copy
 import datetime
@@ -34,10 +37,6 @@ import urllib
 import urlparse
 import uuid
 
-import authomatic.core
-import authomatic.settings as settings
-from authomatic.core import Session
-from authomatic.exceptions import ConfigError, AuthenticationError, FetchError
 
 
 __all__ = ['BaseProvider', 'AuthorizationProvider', 'AuthenticationProvider', 'login_decorator']
@@ -48,7 +47,7 @@ def _error_traceback_html(exc_info, traceback):
     Generates error traceback HTML.
     
     :param tuple exc_info:
-        Output of :finc:`sys.exc_info` function. 
+        Output of :func:`sys.exc_info` function.
         
     :param traceback:
         Output of :func:`traceback.format_exc` function.
@@ -85,11 +84,11 @@ def login_decorator(func):
         try:
             func(provider, *args, **kwargs)
         except Exception as e:
-            if settings.report_errors:
+            if provider.settings.report_errors:
                 error = e
-                provider._log(logging.ERROR, 'Reported supressed exception: {}!'.format(repr(error)))
+                provider._log(logging.ERROR, 'Reported suppressed exception: {0}!'.format(repr(error)))
             else:
-                if settings.debug:
+                if provider.settings.debug:
                     # TODO: Check whether it actually works without middleware
                     provider.write(_error_traceback_html(sys.exc_info(), traceback.format_exc()))
                 raise
@@ -126,10 +125,13 @@ class BaseProvider(object):
     _repr_ignore = ('user',)
     
     __metaclass__ = abc.ABCMeta
+
+    supported_user_attributes = authomatic.core.SupportedUserAttributes()
     
-    def __init__(self, adapter, provider_name, session=None, session_saver=None, callback=None, js_callback=None,
+    def __init__(self, settings, adapter, provider_name, session=None, session_saver=None, callback=None, js_callback=None,
                  prefix='authomatic', **kwargs):
         
+        self.settings = settings
         self.adapter = adapter
         
         self.session = session
@@ -154,10 +156,6 @@ class BaseProvider(object):
         self.popup = self._kwarg(kwargs, 'popup')
     
     
-    ###############################################
-    # From Middleware
-    ###############################################
-    
     @property
     def url(self):
         return self.adapter.url
@@ -268,8 +266,8 @@ class BaseProvider(object):
         """
         
         return kwargs.get(kwname) or \
-               settings.config.get(self.name, {}).get(kwname) or \
-               settings.config.get('__defaults__', {}).get(kwname) or \
+               self.settings.config.get(self.name, {}).get(kwname) or \
+               self.settings.config.get('__defaults__', {}).get(kwname) or \
                default
     
     
@@ -281,7 +279,7 @@ class BaseProvider(object):
             e.g. ``"authomatic:facebook:key"``
         """
         
-        return '{}:{}:{}'.format(settings.prefix, self.name, key)
+        return '{0}:{1}:{2}'.format(self.settings.prefix, self.name, key)
     
     
     def _session_set(self, key, value):
@@ -297,7 +295,7 @@ class BaseProvider(object):
     
     
     @staticmethod
-    def csrf_generator():
+    def csrf_generator(secret):
         """
         Generates CSRF token.
         
@@ -308,7 +306,7 @@ class BaseProvider(object):
         """
         
         # Create hash from random string plus salt.
-        hashed = hashlib.md5(str(uuid.uuid4()) + str(settings.secret)).hexdigest()
+        hashed = hashlib.md5(str(uuid.uuid4()) + str(secret)).hexdigest()
         
         # Each time return random portion of the hash.
         span = 5
@@ -319,7 +317,7 @@ class BaseProvider(object):
     @classmethod
     def _log(cls, level, msg):
         """
-        Logs a message with preformated prefix.
+        Logs a message with pre-formatted prefix.
         
         :param int level:
             Logging level as specified in the
@@ -329,12 +327,12 @@ class BaseProvider(object):
         :param str msg:
             The actual message.
         """
-        
-        authomatic.core._logger.log(level, ': '.join((settings.prefix, cls.__name__, msg)))
-        
+
+        logger = cls._logger or authomatic.core._logger
+        logger.log(level, ': '.join(('authomatic', cls.__name__, msg)))
+
     
-    @classmethod
-    def _fetch(cls, url, method='GET', params=None, headers=None, body='', max_redirects=5, content_parser=None):
+    def _fetch(self, url, method='GET', params=None, headers=None, body='', max_redirects=5, content_parser=None):
         """
         Fetches a URL.
         
@@ -361,15 +359,14 @@ class BaseProvider(object):
         """
         
         params = params or {}
+        params.update(self.access_params)
+        
         headers = headers or {}
+        headers.update(self.access_headers)
         
         scheme, host, path, query, fragment = urlparse.urlsplit(url)
-        
         query = urllib.urlencode(params)
         
-        # Apply headers from settings.
-        headers.update(settings.fetch_headers)
-        
         if method in ('POST', 'PUT', 'PATCH'):
             if not body:
                 # Put querystring to body
@@ -379,12 +376,12 @@ class BaseProvider(object):
         
         request_path = urlparse.urlunsplit((None, None, path, query, None))
         
-        cls._log(logging.DEBUG, u' \u251C\u2500 host: {}'.format(host))
-        cls._log(logging.DEBUG, u' \u251C\u2500 path: {}'.format(request_path))
-        cls._log(logging.DEBUG, u' \u251C\u2500 method: {}'.format(method))
-        cls._log(logging.DEBUG, u' \u251C\u2500 body: {}'.format(body))
-        cls._log(logging.DEBUG, u' \u251C\u2500 params: {}'.format(params))
-        cls._log(logging.DEBUG, u' \u2514\u2500 headers: {}'.format(headers))
+        self._log(logging.DEBUG, u' \u251C\u2500 host: {0}'.format(host))
+        self._log(logging.DEBUG, u' \u251C\u2500 path: {0}'.format(request_path))
+        self._log(logging.DEBUG, u' \u251C\u2500 method: {0}'.format(method))
+        self._log(logging.DEBUG, u' \u251C\u2500 body: {0}'.format(body))
+        self._log(logging.DEBUG, u' \u251C\u2500 params: {0}'.format(params))
+        self._log(logging.DEBUG, u' \u2514\u2500 headers: {0}'.format(headers))
         
         # Connect
         if scheme.lower() == 'https':
@@ -411,11 +408,11 @@ class BaseProvider(object):
             elif max_redirects > 0:
                 remaining_redirects = max_redirects - 1
                 
-                cls._log(logging.DEBUG, 'Redirecting to {}'.format(url))
-                cls._log(logging.DEBUG, 'Remaining redirects: '.format(remaining_redirects))
+                self._log(logging.DEBUG, 'Redirecting to {0}'.format(url))
+                self._log(logging.DEBUG, 'Remaining redirects: '.format(remaining_redirects))
                 
                 # Call this method again.
-                response = cls._fetch(url=location,
+                response = self._fetch(url=location,
                                       params=params,
                                       method=method,
                                       headers=headers,
@@ -426,10 +423,10 @@ class BaseProvider(object):
                                  url=location,
                                  status=response.status)
         else:
-            cls._log(logging.DEBUG, u'Got response:')
-            cls._log(logging.DEBUG, u' \u251C\u2500 url: {}'.format(url))
-            cls._log(logging.DEBUG, u' \u251C\u2500 status: {}'.format(response.status))
-            cls._log(logging.DEBUG, u' \u2514\u2500 headers: {}'.format(response.getheaders()))
+            self._log(logging.DEBUG, u'Got response:')
+            self._log(logging.DEBUG, u' \u251C\u2500 url: {0}'.format(url))
+            self._log(logging.DEBUG, u' \u251C\u2500 status: {0}'.format(response.status))
+            self._log(logging.DEBUG, u' \u2514\u2500 headers: {0}'.format(response.getheaders()))
                 
         return authomatic.core.Response(response, content_parser)
     
@@ -491,6 +488,16 @@ class BaseProvider(object):
         return user
 
 
+    @staticmethod
+    def _http_status_in_category(status, category):
+        """Checks whether a HTTP status code is in the category denoted
+        by the hundreds digit"""
+
+        assert category < 10, 'HTTP status category must be a one-digit int!'
+        cat = category * 100
+        return status >= cat and status < cat + 100
+
+
 class AuthorizationProvider(BaseProvider):
     """
     Base provider for *authorization protocols* i.e. protocols which allow a **provider**
@@ -534,6 +541,16 @@ class AuthorizationProvider(BaseProvider):
             
         :arg dict access_token_params:
             A dictionary of additional request parameters for **access_with_credentials token request**.
+            
+        :arg dict access_headers:
+            A dictionary of default HTTP headers that will be used when
+            accessing **user's** protected resources.
+            Applied by :meth:`.access()`, :meth:`.update_user()` and :meth:`.User.update()`
+            
+        :arg dict access_params:
+            A dictionary of default query string parameters that will be used when
+            accessing **user's** protected resources.
+            Applied by :meth:`.access()`, :meth:`.update_user()` and :meth:`.User.update()`
         """
         
         super(AuthorizationProvider, self).__init__(*args, **kwargs)
@@ -542,14 +559,17 @@ class AuthorizationProvider(BaseProvider):
         self.consumer_secret = self._kwarg(kwargs, 'consumer_secret')
         
         self.user_authorization_params = self._kwarg(kwargs, 'user_authorization_params', {})
-        self.access_token_headers = self._kwarg(kwargs, 'user_authorization_headers', {})
         
+        self.access_token_headers = self._kwarg(kwargs, 'user_authorization_headers', {})
         self.access_token_params = self._kwarg(kwargs, 'access_token_params', {})
         
         self.id = self._kwarg(kwargs, 'id')
         
+        self.access_headers = self._kwarg(kwargs, 'access_headers', {})
+        self.access_params = self._kwarg(kwargs, 'access_params', {})
+        
         #: :class:`.Credentials` to access **user's protected resources**.
-        self.credentials = authomatic.core.Credentials(provider=self)
+        self.credentials = authomatic.core.Credentials(self.settings.config, provider=self)
     
     #===========================================================================
     # Abstract properties
@@ -593,9 +613,9 @@ class AuthorizationProvider(BaseProvider):
         Must convert :data:`credentials` to a :class:`tuple` to be used by :meth:`.Credentials.serialize`.
         
         .. warning::
-        
+
             |classmethod|
-        
+
         :param credentials:
             :class:`.Credentials`
             
@@ -610,9 +630,9 @@ class AuthorizationProvider(BaseProvider):
         Must convert the :data:`deserialized_tuple` back to :class:`.Credentials`.
         
         .. warning::
-        
+
             |classmethod|
-        
+
         :param tuple deserialized_tuple:
             A tuple whose first index is the :attr:`.id` and the rest
             are all the items of the :class:`tuple` created by :meth:`.to_tuple`.
@@ -631,11 +651,11 @@ class AuthorizationProvider(BaseProvider):
         Must return :class:`.RequestElements`.
         
         .. warning::
-        
+
             |classmethod|
-        
+
         :param int request_type:
-            Type of the request specified by one of the classe's constants.
+            Type of the request specified by one of the class's constants.
             
         :param credentials:
             :class:`.Credentials` of the **user** whose
@@ -668,7 +688,7 @@ class AuthorizationProvider(BaseProvider):
     @property
     def type_id(self):
         """
-        A short string reprezenting the provider implementation id used for
+        A short string representing the provider implementation id used for
         serialization of :class:`.Credentials` and to identify the type of provider in JavaScript.
         The part before hyphen denotes the type of the provider, the part after hyphen denotes the class id
         e.g. ``oauth2.Facebook.type_id = '2-5'``, ``oauth1.Twitter.type_id = '1-5'``.
@@ -680,15 +700,13 @@ class AuthorizationProvider(BaseProvider):
         return str(self.PROVIDER_TYPE_ID) + '-' + str(mod.PROVIDER_ID_MAP.index(cls))
     
     
-    @classmethod
-    def access_with_credentials(cls, credentials, url, params=None, method='GET',
-                                headers=None, body='', max_redirects=5, content_parser=None):
+    def access(self, url, params=None, method='GET', headers=None,
+               body='', max_redirects=5, content_parser=None):
         """
-        Fetches the **protected resource** of the **user** to whom belong
-        the supplied :data:`.credentials`.
+        Fetches the **protected resource** of an authenticated **user**.
         
         :param credentials:
-        The **user's** :class:`.Credentials` (serialized or normal).
+            The **user's** :class:`.Credentials` (serialized or normal).
             
         :param str url:
             The URL of the **protected resource**.
@@ -712,64 +730,37 @@ class AuthorizationProvider(BaseProvider):
             :class:`.Response`
         """
         
+        if not self.user and not self.credentials:
+            raise CredentialsError('There is no authenticated user!')
+        
         headers = headers or {}
         
-        cls._log(logging.INFO, 'Accessing protected resource {}.'.format(url))
+        self._log(logging.INFO, 'Accessing protected resource {0}.'.format(url))
         
-        request_elements = cls.create_request_elements(request_type=cls.PROTECTED_RESOURCE_REQUEST_TYPE,
-                                                        credentials=credentials,
+        request_elements = self.create_request_elements(request_type=self.PROTECTED_RESOURCE_REQUEST_TYPE,
+                                                        credentials=self.credentials,
                                                         url=url,
                                                         body=body,
                                                         params=params,
                                                         headers=headers,
                                                         method=method)
         
-        response = cls._fetch(*request_elements,
+        response = self._fetch(*request_elements,
                               max_redirects=max_redirects,
                               content_parser=content_parser)
         
-        cls._log(logging.INFO, 'Got response. HTTP status = {}.'.format(response.status))
+        self._log(logging.INFO, 'Got response. HTTP status = {0}.'.format(response.status))
         return response
-    
-    
-    def access(self, url, params=None, method='GET', headers={}, max_redirects=5, content_parser=None):
-        """
-        Fetches the **protected resource** of the logged in **user**.
-        
-        :param credentials:
-            The **user's** :class:`.Credentials` (serialized or normal).
-            
-        :param str method:
-            HTTP method of the request.
-            
-        :param dict headers:
-            HTTP headers of the request.
-            
-        :param int max_redirects:
-            Maximum number of HTTP redirects to follow.
-            
-        :param function content_parser:
-            A function to be used to parse the :attr:`.Response.data` from :attr:`.Response.content`.
-        
-        :returns:
-            :class:`.Response`
-        """
-        
-        return self.access_with_credentials(credentials=self.credentials,
-                                            url=url,
-                                            params=params,
-                                            method=method,
-                                            headers=headers,
-                                            max_redirects=max_redirects,
-                                            content_parser=content_parser)
 
 
     def async_access(self, *args, **kwargs):
         """
         Same as :meth:`.access` but runs asynchronously in a separate thread.
         
-        .. warning:: |async|
-        
+        .. warning::
+
+            |async|
+
         :returns:
             :class:`.Future` instance representing the separate thread.
         """
@@ -782,9 +773,9 @@ class AuthorizationProvider(BaseProvider):
         Updates the :attr:`.BaseProvider.user`.
         
         .. warning::
-        
+
             Fetches the :attr:`.user_info_url`!
-        
+
         :returns:
             :class:`.UserInfoResponse`
         """
@@ -813,7 +804,7 @@ class AuthorizationProvider(BaseProvider):
         if cls._x_use_authorization_header:
             res = ':'.join((credentials.consumer_key, credentials.consumer_secret))
             res = base64.b64encode(res)
-            return {'Authorization': 'Basic {}'.format(res)}
+            return {'Authorization': 'Basic {0}'.format(res)}
         else:
             return {}
     
@@ -824,10 +815,10 @@ class AuthorizationProvider(BaseProvider):
         """
         
         if not self.consumer.key:
-            raise ConfigError('Consumer key not specified for provider {}!'.format(self.name))
+            raise ConfigError('Consumer key not specified for provider {0}!'.format(self.name))
         
         if not self.consumer.secret:
-            raise ConfigError('Consumer secret not specified for provider {}!'.format(self.name))
+            raise ConfigError('Consumer secret not specified for provider {0}!'.format(self.name))
     
     
     @staticmethod
@@ -846,7 +837,9 @@ class AuthorizationProvider(BaseProvider):
         """
         Override this to handle special request requirements of zealous providers.
         
-        .. warning:: |classmethod|
+        .. warning::
+
+            |classmethod|
         
         :param int request_type:
             Type of request.
@@ -891,7 +884,7 @@ class AuthorizationProvider(BaseProvider):
         
         url = self.user_info_url.format(**self.user.__dict__)
         
-        response = self.access_with_credentials(self.credentials, url)
+        response = self.access(url)
         
         # Create user.
         self.user = self._update_or_create_user(response.data, content=response.content)
@@ -914,11 +907,15 @@ class AuthenticationProvider(BaseProvider):
     def __init__(self, *args, **kwargs):   
         super(AuthenticationProvider, self).__init__(*args, **kwargs)
         
+        # Lookup default identifier, if available in provider
+        default_identifier = getattr(self, 'identifier', None)
+
         # Allow for custom name for the "id" querystring parameter.
         self.identifier_param = kwargs.get('identifier_param', 'id')
         
-        # Get the identifier from request params.
-        self.identifier = self.params.get(self.identifier_param)
+        # Get the identifier from request params, or use default as fallback.
+        self.identifier = self.params.get(
+            self.identifier_param, default_identifier)
         
 
 PROVIDER_ID_MAP = [BaseProvider, AuthorizationProvider, AuthenticationProvider]
